package flowserver

import (
	"encoding/json"
	"errors"
	"flow"
	"flow/registry"
	"fmt"
	"log"
	"reflect"
	"strconv"
	"time"
)

// Node that will contain registry src
type Node struct {
	ID            string            `json:"id"`
	Src           string            `json:"src"`
	Label         string            `json:"label"`
	DefaultInputs map[int]string    `json:"defaultInputs"`
	Prop          map[string]string `json:"prop"`
}

// Link that joins two nodes
type Link struct {
	From string `json:"from"`
	To   string `json:"to"`
	In   int    `json:"in"`
}

// Trigger that join two nodes on state change
type Trigger struct {
	From string   `json:"from"`
	To   string   `json:"to"`
	On   []string `json:"on"`
}

// FlowDocument flow document
type FlowDocument struct {
	Nodes    []Node    `json:"nodes"`
	Links    []Link    `json:"links"`
	Triggers []Trigger `json:"triggers"`
}

func (fd *FlowDocument) fetchNodeByID(ID string) *Node {
	for _, n := range fd.Nodes {
		if n.ID == ID {
			return &n
		}
	}
	return nil
}

func (fd *FlowDocument) fetchTriggerFrom(ID string) []Trigger {
	ret := []Trigger{}
	for _, t := range fd.Triggers {
		if t.From != ID {
			continue
		}
		log.Println("Trigger is:", ID, "adding")
		ret = append(ret, t)
	}
	return ret
}

func (fd *FlowDocument) fetchLinksTo(ID string) []Link {
	ret := []Link{}
	for _, l := range fd.Links {
		if l.To != ID {
			continue
		}
		ret = append(ret, l)
	}
	return ret
}
func (fd *FlowDocument) fetchLinkTo(ID string, n int) *Link {
	for _, l := range fd.Links {
		if l.To != ID || l.In != n {
			continue
		}
		return &l
	}
	return nil
}

var ErrLoop = errors.New("Looping through is disabled for now")

// FlowBuild build the graph based on an starting node
func FlowBuild(rawData []byte, r *registry.R, startingID string) (*flow.Flow, error) {

	doc := FlowDocument{[]Node{}, []Link{}, []Trigger{}}
	log.Println("Building flowDocument from:", string(rawData))
	err := json.Unmarshal(rawData, &doc)
	if err != nil {
		return nil, err
	}

	f := flow.New()
	f.SetRegistry(r)

	nodeTrack := map[string]bool{}

	var build func(ID string) (flow.Operation, error)
	build = func(ID string) (flow.Operation, error) {
		if _, ok := nodeTrack[ID]; ok {
			return nil, ErrLoop //fmt.Errorf("[%v] Looping through nodes is disabled:", ID)
		}
		nodeTrack[ID] = true
		defer delete(nodeTrack, ID)

		// If flow already has ID just return
		if op := f.GetOp(ID); op != nil {
			log.Println("Return operation")
			return op, nil
		}

		node := doc.fetchNodeByID(ID)
		if node == nil {
			return nil, fmt.Errorf("node not found [%v]", startingID)
		}

		var op flow.Operation

		switch node.Src {
		case "Input":
			inputID, err := strconv.Atoi(node.Prop["input"])
			if err != nil {
				return nil, errors.New("Invalid inputID value, must be a number")
			}
			op = f.In(inputID) // By id perhaps
		case "Variable":
			op = f.Var(node.ID, node.Prop["init"])
		case "Const":
			raw := node.Label
			val, err := parseValue(nil, raw)
			if err != nil {
				op, _ = f.DefErrOp(node.ID, err)
			} else {
				op, _ = f.DefConst(node.ID, val)
			}
		default:
			// Load entry
			entry, err := r.Entry(node.Src)
			if err != nil {
				return nil, err
			}
			//// Process inputs ////
			param := make([]flow.Data, len(entry.Inputs))
			for i := range param {
				l := doc.fetchLinkTo(node.ID, i)
				if l == nil {
					// Const value
					v, err := parseValue(entry.Inputs[i], node.DefaultInputs[i])
					if err != nil {
						param[i], _ = f.ErrOp(err)
						continue
					}
					param[i] = v
					continue
				}
				inOp, err := build(l.From)
				if err != nil {
					return nil, err
				}
				param[i] = inOp
			}
			op, err = f.DefOp(node.ID, node.Src, param...)
			if err != nil {
				return nil, err
			}
		}

		// Process triggers for this node
		triggers := doc.fetchTriggerFrom(node.ID)
		for _, t := range triggers {
			_, err := build(t.To)
			if err != nil {
				return nil, err
			}
			// Register the thing here
			f.Hook(flow.Hook{
				Any: func(name string, ID string, triggerTime time.Time, extra ...interface{}) {
					if name != "Error" && name != "Finish" {
						return
					}
					if ID != t.From {
						log.Printf("ID[%v] triggered [%v], I'm t.From: %v", ID, name, t.From)
						return
					}
					exec := false
					for _, o := range t.On {
						if name == o {
							exec = true
							break
						}
					}
					if !exec {
						log.Println("Mismatching trigger, but its a test")
					}

					op := f.Res(t.To)   // Repeating
					go op.Process(name) // Background
				},
			})

		}
		return op, nil
	}

	_, err = build(startingID)
	if err != nil {
		log.Println("Error building")
		return nil, err
	}

	log.Println("Flow:", f)
	return f, nil
}

// Or give a string
func parseValue(typ reflect.Type, raw string) (flow.Data, error) {
	if typ == nil {
		var val flow.Data
		err := json.Unmarshal([]byte(raw), &val)
		if err != nil { // Try to unmarshal as a string?
			val = string(raw)
		}
		return val, nil
	}

	var ret flow.Data
	switch typ.Kind() {
	case reflect.Int:
		v, err := strconv.Atoi(raw)
		if err != nil {
			log.Println("Wrong int conversion", err)
			return nil, err
		}
		ret = v
	case reflect.String:
		ret = raw
	default:
		if len(raw) == 0 {
			ret = reflect.Zero(typ)
		} else {
			refVal := reflect.New(typ)
			err := json.Unmarshal([]byte(raw), refVal.Interface())
			if err != nil {
				return nil, err
			}
			ret = refVal.Interface()
		}
	}
	log.Printf("Returning %#v", ret)
	return ret, nil
}
